/******************************************************************************
 *
 * mynumfld.m
 *
 * date:   10/9/2002
 * author: Nils Bruin
 *
 * Some features missing for number field and related objects
 *
 * Included LocalRing from RngLoc/order.m by Claus Fieker to fix a precision
 * bug.
 ******************************************************************************/

/* We'll be saving information in prime ideals later on. Therefore, it is good
to be able to get a "standard version" of a prime ideal.  With LookupPrime,
we store & retrieve standard versions of a prime ideal. Be sure to call

p:=LookupPrime(p);

whenever you want to use the functions here in order to get higher consistency
in returned values and a better performance */

declare attributes RngOrd: StandardPrimeIdeals;

intrinsic LookupPrime(P::RngOrdIdl) -> RngOrdIdl
  {Routine to get *IDENTICAL* prime ideals for a number field.
  Calling this routine several times with equal prime ideals of the same order
  is guaranteed to return identical prime ideals. This enables attributes to
  be stored in prime ideals and preserved along different instantiations
  of that prime ideal}
  O:=Order(P);
  if not(assigned O`StandardPrimeIdeals) then
    O`StandardPrimeIdeals:={};
  end if;
  if exists(Q){R:R in O`StandardPrimeIdeals| P eq R} then
    return Q;
  else
    Include(~O`StandardPrimeIdeals,P);
    return P;
  end if;
end intrinsic;

IdealRecordFormat:=recformat<LocalTwoSelmerMap,SafeUniformiser,
Completion,ComplMap>;

/**
 ** Completion with memory
 **/

intrinsic MyCompletion(p::RngOrdIdl:Precision:=false)->FldLoc,Map
  {Returns a completion with memory. If precision is given and is different from
  a stored completion, then stored field will be changed (recreated)}
  if not(assigned p`Record) then
    p`Record:=rec<IdealRecordFormat|>;
  end if;
  if not(assigned p`Record`Completion) or Precision cmpne false then
    if Precision cmpne false then
      Kp,mp:=Completion(NumberField(Order(p)),p:Precision:=Precision);
    else
      Kp,mp:=Completion(NumberField(Order(p)),p);
    end if;
    rc:=p`Record;
    rc`Completion:=Kp;
    rc`ComplMap:=map<Domain(mp)->Codomain(mp)|
          a:->mp(a*Denominator(a))/mp(Denominator(a)),
	  b:->Inverse(mp)(b)>;
    p`Record:=rc;
  end if;
  return p`Record`Completion,p`Record`ComplMap;
end intrinsic;

/**
 ** SafeUniformiser
 **
 ** retrieve or compute and store an element of valuation 1 at the given prime
 ** and valuation 0 at the other primes above the same rational prime.
 **/

intrinsic SafeUniformiser(p::RngOrdIdl)->RngOrdElt
  {returns an element of valuation 1 at the given prime and of
   valuation 0 at the other primes over Norm(p)}
   if not(assigned p`Record) then
     p`Record:=rec<IdealRecordFormat|>;
   end if;
   if not(assigned p`Record`SafeUniformiser) then
     I:=[q[1]: q in Factorisation(Norm(p)*Order(p))| q[1] ne p];
     u:=WeakApproximation([p]cat I,[1] cat [0: i in I]);
     rc:=p`Record;
     rc`SafeUniformiser:=u;
     p`Record:=rc;
   end if;
   return p`Record`SafeUniformiser;
end intrinsic;

/**
 ** LocalTwoSelmerMap
 **
 ** This routine returns a map from K^* -> Kp^* /(Kp^*2). If called twice with
 ** the same prime, returns the identical map. If called on equal but not
 ** identical primes, returns different maps that may not even agree
 ** coordinate-wise.
 **/
 
intrinsic LocalTwoSelmerMap(p::RngOrdIdl)->Map
  {returns a map from K^* to the group Kp^*/(Kp^*2)}
  if not(assigned p`Record) then
    p`Record:=rec<IdealRecordFormat|>;
  end if;
  if not(assigned p`Record`LocalTwoSelmerMap) then
    // We assume that grp1 is returned in Smith Normal Form!!!
    // therefore, by construction, we know that grp1 has all its
    // factors of even order.
    grp1,mp1i:=RayResidueRing(p^(Valuation(Order(p)!4,p)+1));
    grp2,mp2:=quo<grp1|[2*grp1.i: i in [1..Ngens(grp1)]]>;
    mp1:=Inverse(mp1i);
    mp2i:=Inverse(mp2);
    u:=SafeUniformiser(p);
    grp:=AbelianGroup([2:i in [1..#Invariants(grp1)+1]]);
    fn:=function(a)
	if (a eq 0) then
          error "LocalTwoSelmerMap only defined on multiplicative group";
	else
          a:=a*Denominator(a)^2;
	  v:=Valuation(a,p);
	  a:=a/u^v;
          return grp!([v]cat Eltseq(mp2(mp1(a))));
 	end if;
      end function;
    fni:=function(b)
        v:=Eltseq(b);
        return u^v[1]*mp1i(mp2i(grp2![v[i]:i in [2..#Invariants(grp1)+1]]));
      end function;
    grpmap:=pmap<NumberField(Order(p))->grp|a:->fn(a),b:->fni(b)>;
    rc:=p`Record;
    rc`LocalTwoSelmerMap:=grpmap;
    p`Record:=rc;
  end if;
  return p`Record`LocalTwoSelmerMap;
end intrinsic;

/**
 ** Support(RngOrdIdl)
 **
 ** factorisation without multiplicity.
 **/

intrinsic Support(I::RngOrdIdl) -> SetEnum
  {returns the prime divisors of an ideal as a set}
  return {p[1]:p in Factorisation(I)};
end intrinsic;

intrinsic MyFactorisation(f::RngUPolElt) -> SeqEnum
  {Stupid factorisation that only splits of linear factors. Complains if
   cofactor is of degree larger than 3}
  require LeadingCoefficient(f) eq 1 : "Can only handle monic polynomials";
  R:=Parent(f);
  if Type(BaseRing(f)) eq RngLoc then
    rts:=IntegralRoots(f);
  else
    rts:=MyRoots(f);
  end if;
  res:=[<R.1-r[1],r[2]>:r in rts];
  cofact:=&*[r[1]^r[2]:r in res];
  if Degree(f) gt Degree(cofact)+3 then
    print "WARNING: only linear factors split off";
  end if;
  return res cat [<f div cofact,1>];
end intrinsic;

//Currently, due to trouble with p-adic factorisation
//(precision choice in LocalRing versus what Factorisation thinks is acceptable)
//we include a routine from RngLoc/order.m here that is due to Claus Fieker.
//We slightly increase the used precision here to get rid of nasty error
//messages. This should not be necessary in later versions anymore.

do_old := false;

intrinsic LocalRing(P::RngOrdIdl, Prec::RngIntElt) -> RngLoc, Map
{the completion at P up to a precision of Prec together with the embedding}
  O := Order(P);

  require IsAbsoluteOrder(O) : "The ideal must belong to some absolue order";
  require IsPrime(P) : "The ideal must be prime";
  require IsMaximal(O) or RamificationDegree(P) eq 1 : "For ramified primes the order must be maximal";
  require IsSimple(O) : "The order must be simple";

  Z := BaseRing(O);
  f := InertiaDegree(P);
  e := RamificationDegree(P);
  p := Minimum(P);

  Lift := function(f, w, P, k)
    pk := P^k;
    O := Order(P);
    fs := Derivative(f);

    modP, qP := quo<O | P>;
    P := P^2;
    mod2P, q2P := quo<O | P>;

    b := (1/qP(Evaluate(fs, w))) @@ qP;
    w := q2P(w);
    w := (w - q2P(b)*Evaluate(PolynomialRing(mod2P)!f, w)) @@ q2P; 
    k := k div 2;
    while k gt 0 do
      modP := mod2P;
      qP := q2P;
      P := P^2;
      k := k div 2;

      mod2P, q2P := quo<O | P>;

      b := (qP(b)*(2-qP(b)*Evaluate(fs, qP(w))))  @@ qP;
      w := (q2P(w)-q2P(b)*Evaluate(f, q2P(w))) @@ q2P;
    end while;
    
    _, qP := quo<O | pk>;
    return (qP(w)) @@ qP;
  end function;  


  if exists{x: x in Decomposition(O, p) | x[2] gt 1} or
    Valuation(LeadingCoefficient(DefiningPolynomial(O)), p) ne 0
     or Valuation(Index(O, EquationOrder(O)), p) gt 0 then

    if do_old then
      // get the minimal precision and use lifting for the other part:
      prec := 3*Valuation(ReducedDiscriminant(EquationOrder(O)), p)+1;
      // 2 should be enough, but currently this does not work!!
      require prec le Prec : "Use at least a precision of ", prec;
    else
      lc := LeadingCoefficient(DefiningPolynomial(O));
      if IsOne(lc) then
        prec := Valuation(Discriminant(EquationOrder(O)), p)+1;
        do_ring := true;
      else
        ff := Evaluate(DefiningPolynomial(O), 
                     PolynomialRing(Rationals()).1/lc)*lc^(Degree(O)-1);
        ff := Polynomial(Integers(), ff);             
        prec := Valuation(Discriminant(ff), p)+1;
        do_ring := false;
      end if;
      // get a useful estimate!!!
    end if;
    PRec := Prec + e*prec;

    // set up the unramified part:

    Fq, q := ResidueClassField(P);
    w := Generator(Fq, PrimeField(Fq));
    w := w @@ q;   // now w should be a generator for the unramified part
    if f gt 1 then

      M := MinimalPolynomial(q(w));
      M := PolynomialRing(Z)!M;

      Qf1 := LocalRing(p, M, Prec);
      // OK, now we have to lift w in O to get the embedding:
      w := Lift(M, w, P, PRec);
      Qf := LocalRing(p, M, PRec);
    else
      Qf := LocalRing(p, 1, PRec);
      Qf1 := LocalRing(p, 1, prec);
    end if;  

    N := O!1;
    M := [O|];
    for j in [0..e-1] do
      if f gt 1 then
        MM := [O!1*N, w*N];
        for i in [2..f-1] do
          Append(~MM, MM[i] * w);
        end for;
      else
        MM := [O!1*N];
      end if;  
      Reverse(~MM);
      M := M cat MM;
      N := N*UniformizingElement(P);
    end for;
    Append(~M, N);
    Reverse(~M);

    Ppow := BasisMatrix(P^(e*PRec+1));
    M := [ [Z!y : y in Eltseq(x)] : x in M] cat
         [  Eltseq(Ppow[x]) : x in [1..Degree(O)] ];
    M := Matrix(M);     
    M := KernelMatrix(M);
    M := EchelonForm(M);

    M := Eltseq(M[1]);

    M := [ Z!(M[x]/M[1]) : x in [1..#M-Degree(O)]];

    N := [Qf!1];
    for j in [0..e-1] do
      Append(~N, Qf!(M[2+f*j..1+f*(j+1)]));
    end for;  
    Reverse(~N);
    
    N := PolynomialRing(Qf)!N;
    if do_ring then
      Qfe := LocalRing(Qf, N, e*PRec);
      N := PolynomialRing(Qf1)!N;
      Qf1e := LocalRing(Qf1, N, e*prec);
    else
      Qfe := LocalField(FieldOfFractions(Qf), N, e*PRec);
      N := PolynomialRing(FieldOfFractions(Qf1))!N;
      Qf1e := LocalField(FieldOfFractions(Qf1), N, e*prec);
    end if;
//    SuggestedPrecision(PolynomialRing(Qf1e)! DefiningPolynomial(O));e*prec;
    if do_old then
      F := Factorisation(PolynomialRing(Qf1e)! DefiningPolynomial(O));
    
      F := [x : x in F | Degree(x[1]) eq 1];

      z := [ Valuation(Qf1e.1+Qf1e.2 - hom<O ->Qf1e | -Eltseq(x[1])[1]>(
                         UniformizingElement(P)+w
                                                                )
                       ) : x in F | Degree(x[1]) eq 1 ];
     
      _, i := Maximum(z);
      
      F := - Eltseq(F[i][1])[1];
    else
// instead of factorisation (which is chancy currently, do linear algebra...)
      K := NumberField(O);
      b := [Eltseq(O!w^i*UniformizingElement(P)^j) : i in [0..f-1], j in [0..e-1]];
      b := Matrix(Integers(), Matrix(b));
      b := VerticalJoin(b, BasisMatrix(P^(prec)));
      // b must be extended by a appropriate prime ideal power...
      pe := FieldOfFractions(O)!K.1;
      d := Denominator(pe);
      s := Solution(b, Matrix(Integers(), [Eltseq(d*pe)]));
      s := Eltseq(s);
      F := Qf1e![Qf1!s[(i-1)*f+1..i*f] : i in [1..e]]/d;
//      Valuation(Evaluate(DefiningPolynomial(O), F));
//      Valuation(Evaluate(Derivative(DefiningPolynomial(O)), F));
    end if;
    F := HenselLift(PolynomialRing(Qfe)!DefiningPolynomial(O), Qfe!F, Prec*e);

    ChangePrecision(~Qfe, Prec*e);
    F := Qfe!F;

    h := hom<O -> FieldOfFractions(Qfe) | F>;
    h := map<O -> Qfe | x:->Qfe!h(x)>;
  else
    Qf1 := LocalRing(p, 1, 2);
    require 2 le Prec : "Use at least a precision of ", 2;

    alpha := O!PrimitiveElement(EquationOrder(O));

    G := Factorisation(PolynomialRing(Qf1)!DefiningPolynomial(O));
    
    F := [ G[x] : x in 
            [ y : y in [1..#G] | Degree(G[y][1]) eq f and G[y][2] eq 1 ] ] ;
    F := [ <[ Z!x : x in Eltseq(y[1]) ], y[1]>  : y in F ];
    F := [ < PolynomialRing(Z)!x[1], x[2]>  : x in F];
    F := [ <Evaluate(x[1], alpha), x[2]> : x in F];

    F := [ x : x in F| x[1] in P];
    assert #F eq 1;

    F1 := F[1][2];
    if #G eq 1 then
      F2 := Parent(F1)!1;
    else  
      F2 := &* [x[1] : x in G | x[1] ne F1];
    end if;     

    Qf := LocalRing(p, 1, Prec);
    F1, _ := HenselLift(PolynomialRing(Qf)!DefiningPolynomial(O), F1, F2);
  
    if f eq 1 then
      Qfe := LocalRing(p, 1, Prec);
      h := hom<O -> Qfe | -Eltseq(F1)[1] >;
    else
      F1 := [ Z!x : x in Eltseq(F1)];
      F1 := PolynomialRing(Z)!F1;

      Qfe := LocalRing(p, F1, Prec);
      h := hom<O -> Qfe | Qfe.2>;
    end if;  
    w := O!PrimitiveElement(EquationOrder(O));
  end if;

  // embeddings:
  // O -> Local


  HINV := function(x)
    if e eq 1 then
      l := [Z! i : i in InertseqpAdic(x)];
      l := PolynomialRing(Z)!l;
      return Evaluate(l, w);
    else
      l := Locseq(x);
      l := [ Evaluate(PolynomialRing(Z)![ Z!x : x in s], w) : s in l];
      return Evaluate(PolynomialRing(O)!l, UniformizingElement(P));
    end if;
  end function;

  return Qfe, map<O -> Qfe | x:->h(x), y:-> HINV(y)>;

end intrinsic;
