/******************************************************************************
 *
 * newell.m
 *
 * date:   14/8/2001
 * author: Nils Bruin
 *
 * Some routines for added functionality for elliptic curves over number fields
 *
 ******************************************************************************/

/**
 ** EllIsIsomorphic
 **
 ** Extension of IsIsomorphic for elliptic curves. Present standard
 ** implementation fails for curves not over finite fields or Q. This should
 ** be changed. 
 **/
 
intrinsic EllIsIsomorphic(ec1::CrvEll,ec2::CrvEll) -> BoolElt, Map
  {Tests if ec1 and ec2 are isomorphic over their common base field of
  characteristic prime to 6. If they are, the isomorphism is
  returned}
  require BaseRing(ec1) cmpeq BaseRing(ec2): 
    "Elliptic curves must be over same base ring";
  a:=aInvariants(ec2);
  b:=aInvariants(ec1);
  P:=PolynomialAlgebra(BaseRing(ec1));
  s4val:=Roots(Gcd(-8*b[1]^2*s4^4*b[2]+24*b[1]*s4^4*b[3]-16*b[2]^2*s4^4-24*a[1]*a[3
]-b[1]^4*s4^4+16*a[2]^2+8*a[1]^2*a[2]+48*b[4]*s4^4+a[1]^4-48*a[4], -128*a[2]^3
-432*a[3]^2+72*a[3]*a[1]^3+288*a[3]*a[1]*a[2]+576*a[4]*a[2]+144*a[4]*a[1]^2-24
*a[2]*a[1]^4-96*a[1]^2*a[2]^2-2*a[1]^6+48*b[1]^2*s4^2*a[2]^2+192*a[2]^2*b[2]*
s4^2+3*a[1]^4*b[1]^2*s4^2+12*a[1]^4*b[2]*s4^2-12*b[1]^4*s4^6*b[2]-48*b[1]^2*s4
^6*b[2]^2+1728*b[5]*s4^6-64*b[2]^3*s4^6+432*b[3]^2*s4^6-576*a[4]*b[2]*s4^2+24*
a[2]*a[1]^2*b[1]^2*s4^2+96*a[2]*a[1]^2*b[2]*s4^2-72*a[3]*a[1]*b[1]^2*s4^2-288*
a[3]*a[1]*b[2]*s4^2-144*a[4]*b[1]^2*s4^2-b[1]^6*s4^6-1728*a[5])) where s4:=P.1;
  if #s4val eq 0 then
    return false,_;
  else
    return true,Isomorphism(ec1,ec2,[-1/12*a[1]^2+1/12*b[1]^2*s4^2-1/3*a[2]+1/3*b[2]*s4^2, -1/2*a[1]
+1/2*b[1]*s4, -1/2*a[3]+1/24*a[1]^3-1/24*a[1]*b[1]^2*s4^2+1/6*a[1]*a[2]-1/6*a[
1]*b[2]*s4^2+1/2*b[3]*s4^3, s4]) where s4 := s4val[1][1];
  end if;
end intrinsic;

/**
 ** EllSymmetricModel
 **
 ** Changes the model of an elliptic curve to y^2=x^3+a2*x^2+a4*x+a6
 ** and preserves integrality of the model, if appropriate.
 **/
intrinsic EllSymmetricModel(ec::CrvEll) -> CrvEll, Map
  {returns a model of the form y^2=x^3+a2*x^2+a4*x+a6 together with
  isomorphisms. Will not work in characteristic 2.}
  a:=aInvariants(ec);
  ecr, f1 := Transformation(ec,[0,a[1]/2,a[3]/2,1]);
  if (Category(BaseRing(ec)) eq FldRat) or
    (Category(BaseRing(ec)) eq FldNum) and
    EllIsIntegral(ec) and not(EllIsIntegral(ecr)) then
      ecr, f1 := Transformation(ec,[0,a[1],a[3]*4,2]);
  end if;
  return ecr, f1;
end intrinsic;

/**
 ** EllIsIntegral
 **
 ** tests whether an elliptic curve over the rationals or over a number field
 ** has integral coefficients.
 **/
 
intrinsic EllIsIntegral(ec::CrvEll) -> BoolElt
  {returns true if an elliptic curve over a number field
  has integral coefficients}
  require Category(BaseRing(ec)) eq FldRat or Category(BaseRing(ec)) eq FldNum:
  "Elliptic curve should be over a number field";
  return forall{a:a in aInvariants(ec)|IsIntegral(a)};
end intrinsic; 

intrinsic IsIntegral(E::CrvEll,p::RngOrdIdl) ->BoolElt
  {returns true if the defining coefficients of E have non-negative valuation at
  the prime ideal p}
  cfs:=Eltseq(E);
  return forall{c:c in cfs|Valuation(c,p) ge 0};
end intrinsic;

/**
 ** EllIntegralModel
 **
 ** Changes the model of an elliptic curve over the rationals or over a number
 ** field to one with integral coefficients.
 ** it tries to do so in a carefull way, but the returned model is not
 ** necessarily minimal in the standard sence.
 **/
 
intrinsic EllIntegralModel(ec::CrvEll) -> CrvEll, Map, Map
  {returns a model with integral coefficients of an elliptic curve over a number
  field}
  require Category(BaseRing(ec)) eq FldRat or Category(BaseRing(ec)) eq FldNum:
  "Elliptic curve should be over a number field";
  dn:=1; ecn:=ec; phi:=IdentityMap(ec); phiinv:=phi;
  while not forall(elt){Denominator(a): a in aInvariants(ecn)| IsIntegral(a)} do
    dn*:=elt;
    ecn,phi:=Transformation(ec,[0,0,0,dn]);
  end while;
  return ecn,phi,phiinv;
end intrinsic;

/**
 ** UnivariatePolynomial(CrvEll)
 **
 ** Returns the 2-torsion polynomial for a curve in the forn y^2=F(x)
 **/

intrinsic UnivariatePolynomial(E::CrvEll)->RngUPolElt
  {Provided E is of the form y^2=F(x)=x^3+a2*x^2+a4*x+a6, return F}
  a:=aInvariants(E);
  require a[1] eq 0 and a[3] eq 0 :
    "Only makes sense for [0,a2,0,a4,a6] curves";
  return Polynomial(BaseRing(E),[a[5],a[4],a[2],1]);
end intrinsic;

/**
 ** PolToEllipticCurve
 **
 ** Takes a square free degree 3 polynomial F and creates the elliptic curve
 ** y^2=F(x). F need not be monic, so a great way for creating twists.
 **/

intrinsic PolToEllipticCurve(F::RngUPolElt) -> CrvEll
  { returns the elliptic curve y^2=F(x). F should be square-free and of degree
  3. For F=f0*x^3+f1*x^2+f2*x+f3, the curve
  [0,f1,0,f2*f0,f3*f0^2] is returned.}
  require Degree(F) eq 3 and Discriminant(F) ne 0:
    "Only defined for square-free degree 3 polynomials";
  return EllipticCurve([0,f[3],0,f[2]*f[4],f[1]*f[4]^2]) where f:=Eltseq(F);
end intrinsic;

declare attributes Map:Isogeny_Dual;

intrinsic DualIsogeny(phi::Map)->Map
  {gives dual isogeny}
  if not(assigned phi`Isogeny_Dual) then
    dual:=IsogenyFromKernel(
       PushThroughIsogeny(phi,Kernel(MultiplicationByMMap(Domain(phi),
       Degree(phi)))));
    bool,isom:=EllIsIsomorphic(Image(dual),Domain(phi));
    error if not bool, "Unexpected failure in computing dual isogeny";
    dual:=dual*isom;
    phi`Isogeny_Dual:=dual;
    dual`Isogeny_Dual:=phi;
  end if;
  return phi`Isogeny_Dual;  
end intrinsic;

intrinsic TwoIsogeny(P::PtEll)->Map
  {computes the 2-isogeny with the 2-torsion point P in the kernel}
  E:=Curve(P);
  a:=aInvariants(E);
  require (a[1] eq 0) and (a[3] eq 0):
    "Elliptic curve must be of form [0,a2,0,a4,a6]";
  require (P[2] eq 0) and (P[3] ne 0):
    "Point P must be a 2-torsion point";
  E1,mp1:=Transformation(E,[-P[1]/P[3],0,0,1]);
  mp2:=Inverse(mp1);
  P<x,y>:=CoordinateRing(Ambient(AffinePatch(E1,1)));
  a:=aInvariants(E1);
  b4:=a[2]^2-4*a[4];
  E2:=EllipticCurve([0,-2*a[2],0,b4,0]);
  psi1:=mp1*Isogeny(E1,E2,x,x^2+a[2]*x+a[4],y*(a[4]-x^2));
  psi2:=Isogeny(E2,E1,2*x,x^2-2*a[2]*x+b4,y*(a[4]-x^2))*mp2;
  psi1`Isogeny_Dual:=psi2;
  psi2`Isogeny_Dual:=psi1;
  return psi1;
end intrinsic;

intrinsic IsogenyIsMultiplicationByTwo(phi::Map)->BoolElt
  {tests if an isogeny of elliptic curves is multiplication by 2}
  return (Domain(phi) cmpeq Codomain(phi)) and
         (IsogenyMapPsi(phi) eq UnivariatePolynomial(Codomain(phi)));
end intrinsic;

intrinsic Reduction(E::CrvEll,p::RngOrdIdl)->CrvEll,Map
  {Returns the reduction of an elliptic curve at a prime ideal, provided the
  given model has good reduction at the prime. The map gives the reduction map
  on the Mordell-Weil group of the curve. Obviously, there is no inverse.}
  K:=BaseRing(E);
  if Type(K) eq RngOrd then
    K:=NumberField(K);
  end if;
  require (Type(K) eq FldNum and NumberField(Order(p)) eq K) or
          (Type(K) eq RngOrd and Order(p) eq K):
    "Field of definition and ideal are not compatible";
  cfs:=Eltseq(E);
  require IsIntegral(E,p) and
        Valuation(Discriminant(E),p) eq 0:
    "model should be integral and of good reduction at the prime";
  Kbar,red:=ResidueClassField(IntegerRing(K),p);
  Ebar:=EllipticCurve([red(c):c in cfs]);
  u:=SafeUniformiser(p);
  fn:=function(P)
    L:=Eltseq(P);
    dn:=&*[Denominator(c):c in L];
    L:=[dn*c:c in L];
    v:=Minimum([Valuation(c,p):c in L|c ne 0]);
    return Ebar![red(c/u^v):c in L];
  end function;
  return Ebar,map<E->Ebar| P:-> fn(P)>;
end intrinsic;

intrinsic TorsionBound(E::CrvEll,n::RngIntElt) -> RngIntElt
  {Returns a bound on the size of the torsion subgroup of E by looking at
  the first n non-inert primes of sufficiently low ramification degree of
  good reduction}
  Bound:=0;
  try:=0;
  Dsc:=Discriminant(E);
  K:=BaseRing(E);
  O:=IntegerRing(K);
  p:=1;
  repeat
    p:=NextPrime(p:Proof:=false);
    S:=Support(p*O);
    for P in S do
      if InertiaDegree(P) eq 1 and RamificationDegree(P)/(Minimum(P)-1) lt 1
           and IsIntegral(E,P) and Valuation(Dsc,P) eq 0 then
        Bound:=GCD(Bound,#Reduction(E,P));
	try+:=1;
	if try ge n then
	  return Bound;
	end if;
      end if;
    end for;
  until false;
end intrinsic;

intrinsic pPowerTorsion(E::CrvEll,p::RngIntElt:Bound:=-1)->SeqEnum
  {returns the pPower torsion on E. The optional parameter gives an upper bound
  on the size of the p-power torsion subgroup that is used in the computations.}
  require IsPrime(p):"p should be a prime";
  a:=0;b:=0;
  L:=[];
  defpol:=DefiningPolynomial(AffinePatch(E,1));
  if p eq 2 then
    Fn:=DivisionPolynomial(E,p);
    n:=1;
    L[1]:=[x[1]:x in Roots(Fn)|HasRoot(Evaluate(defpol,[x[1],Parent(Fn).1]))];
    if #L[1] ge 1 then
      a+:=1;
    end if;
    if #L[1] eq 3 then
      b+:=1;
    end if;
  else
    n:=0;
    Fn:=1;
  end if;
  while n eq 0 or (#L[n] gt 0 and (Bound lt 0 or p^(a+b) lt Bound)) do
    Fnm1:=Fn;
    n:=n+1;
    Fn:=DivisionPolynomial(E,p^n);
    L[n]:=[x[1]:x in Roots(Parent(Fn)!(Fn/Fnm1))
        |HasRoot(Evaluate(defpol,[x[1],Parent(Fn).1]))];
   if #L[n] ge (p^(n-1)*(p-1))/2 then
      a+:=1;
    end if;
    if #L[n] ge (p^(2*n-2)*(p^2-1))/2 then
      b+:=b;
    end if;
  end while;
  result:=[];
  if a ge 1 then
    T1:=E![L[a][1],Roots(Evaluate(DefiningPolynomial(AffinePatch(E,1)),
                           [L[a][1],Parent(Fn).1]))[1][1]];
    Append(~result,<T1,p^a>);
  end if;
  if b ge 1 then
    pT1:=p^(b-a)*T1;
    xcoors:={(c*pT1)[1]:c in [1..p^(b-1)*(p-1)]| c mod p ne 0};
    error if not exists(T2x){x: x in L[b]| not(x in xcoors)},
      "can't find second generator for torsion";
    T2:=E![T2x,Roots(Evaluate(DefiningPolynomial(AffinePatch(E,1)),
                         [T2x,Parent(Fn).1]))[1][1]];
    Append(~result,<T2,p^b>);
  end if;
  return result;
end intrinsic;

intrinsic MyTorsionSubgroup(E::CrvEll)->GrpAb, Map
  {Computes the 2-torsion subgroup}
  T1:=E!0;ordT1:=1;
  T2:=E!0;ordT2:=1;
  for tup in Factorisation(TorsionBound(E,20)) do
    tor:=pPowerTorsion(E,tup[1]:Bound:=tup[1]^tup[2]);
    if #tor ge 1 then
      T1+:=tor[1][1];
      ordT1*:=tor[1][2];
    end if;
    if #tor eq 2 then
      T2+:=tor[2][1];
      ordT2*:=tor[2][2];
    end if;
  end for;
  if ordT1 eq 1 then
    grp:=AbelianGroup([1]);
    mp:=map<grp->E|a:->E!0>;
  elif ordT2 eq 1 then
    grp:=AbelianGroup([ordT1]);
    mp:=map<grp->E|a:->Eltseq(a)[1]*T1>;
  else
    grp:=AbelianGroup([ordT1,ordT2]);
    mp:=map<grp->E|a:->Eltseq(a)[1]*T1+Eltseq(a)[2]*T2>;
  end if;
  return grp,mp;
end intrinsic;
